@@ -12,6 +12,9 @@
 
 #define XDG_OUTPUT_MANAGER_VERSION 3
 
+#define LINUX_DMABUF_VERSION 4
+#define LINUX_DMABUF_VERSION_MIN 3
+
 struct xdpw_state;
 
 int xdpw_wlr_screencopy_init(struct xdpw_state *state);
@@ -24,7 +27,8 @@ struct xdpw_wlr_output *xdpw_wlr_output_find(struct xdpw_screencast_context *ctx
 	struct wl_output *out, uint32_t id);
 struct xdpw_wlr_output *xdpw_wlr_output_chooser(struct xdpw_screencast_context *ctx);
 
-void xdpw_wlr_frame_free(struct xdpw_screencast_instance *cast);
+void xdpw_wlr_frame_finish(struct xdpw_screencast_instance *cast);
+void xdpw_wlr_frame_start(struct xdpw_screencast_instance *cast);
 void xdpw_wlr_register_cb(struct xdpw_screencast_instance *cast);
 
 #endif
