@@ -1,5 +1,6 @@
 #include "wlr_screencast.h"
 
+#include "linux-dmabuf-unstable-v1-client-protocol.h"
 #include "wlr-screencopy-unstable-v1-client-protocol.h"
 #include "xdg-output-unstable-v1-client-protocol.h"
 #include <fcntl.h>
@@ -12,7 +13,9 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#include <assert.h>
 #include <wayland-client-protocol.h>
+#include <xf86drm.h>
 
 #include "screencast.h"
 #include "pipewire_screencast.h"
@@ -20,196 +23,219 @@
 #include "logger.h"
 #include "fps_limit.h"
 
-static void wlr_frame_buffer_destroy(struct xdpw_screencast_instance *cast) {
-	// Even though this check may be deemed unnecessary,
-	// this has been found to cause SEGFAULTs, like this one:
-	// https://github.com/emersion/xdg-desktop-portal-wlr/issues/50
-	if (cast->simple_frame.data != NULL) {
-		munmap(cast->simple_frame.data, cast->simple_frame.size);
-		cast->simple_frame.data = NULL;
+void wlr_frame_free(struct xdpw_screencast_instance *cast) {
+	if (!cast->wlr_frame) {
+		return;
 	}
-
-	if (cast->simple_frame.buffer != NULL) {
-		wl_buffer_destroy(cast->simple_frame.buffer);
-		cast->simple_frame.buffer = NULL;
-	}
-}
-
-void xdpw_wlr_frame_free(struct xdpw_screencast_instance *cast) {
 	zwlr_screencopy_frame_v1_destroy(cast->wlr_frame);
 	cast->wlr_frame = NULL;
-	if (cast->quit || cast->err) {
-		wlr_frame_buffer_destroy(cast);
-		logprint(TRACE, "xdpw: simple_frame buffer destroyed");
-	}
 	logprint(TRACE, "wlroots: frame destroyed");
+}
+
+void xdpw_wlr_frame_finish(struct xdpw_screencast_instance *cast) {
+	logprint(TRACE, "wlroots: finish screencopy");
+
+	wlr_frame_free(cast);
 
 	if (cast->quit || cast->err) {
 		// TODO: revisit the exit condition (remove quit?)
 		// and clean up sessions that still exist if err
 		// is the cause of the instance_destroy call
 		xdpw_screencast_instance_destroy(cast);
-		return ;
+		return;
 	}
 
-	uint64_t delay_ns = fps_limit_measure_end(&cast->fps_limit, cast->framerate);
-	if (delay_ns > 0) {
-		xdpw_add_timer(cast->ctx->state, delay_ns,
-			(xdpw_event_loop_timer_func_t) xdpw_wlr_register_cb, cast);
-	} else {
-		xdpw_wlr_register_cb(cast);
-	}
-}
-
-static struct wl_buffer *create_shm_buffer(struct xdpw_screencast_instance *cast,
-		enum wl_shm_format fmt, int width, int height, int stride,
-		void **data_out) {
-	struct xdpw_screencast_context *ctx = cast->ctx;
-	int size = stride * height;
-
-	int fd = anonymous_shm_open();
-	if (fd < 0) {
-		logprint(ERROR, "wlroots: shm_open failed");
-		return NULL;
+	if (!cast->pwr_stream_state) {
+		cast->frame_state = XDPW_FRAME_STATE_NONE;
+		return;
 	}
 
-	int ret;
-	while ((ret = ftruncate(fd, size)) == EINTR);
-
-	if (ret < 0) {
-		close(fd);
-		logprint(ERROR, "wlroots: ftruncate failed");
-		return NULL;
-	}
-
-	void *data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
-	if (data == MAP_FAILED) {
-		logprint(ERROR, "wlroots: mmap failed: %m");
-		close(fd);
-		return NULL;
-	}
-
-	struct wl_shm_pool *pool = wl_shm_create_pool(ctx->shm, fd, size);
-	close(fd);
-	struct wl_buffer *buffer =
-		wl_shm_pool_create_buffer(pool, 0, width, height, stride, fmt);
-	wl_shm_pool_destroy(pool);
-
-	*data_out = data;
-	return buffer;
-}
-
-static void wlr_frame_buffer_chparam(struct xdpw_screencast_instance *cast,
-		uint32_t format, uint32_t width, uint32_t height, uint32_t stride) {
-	logprint(DEBUG, "wlroots: reset buffer");
-	cast->simple_frame.width = width;
-	cast->simple_frame.height = height;
-	cast->simple_frame.stride = stride;
-	cast->simple_frame.size = stride * height;
-	cast->simple_frame.format = format;
-	wlr_frame_buffer_destroy(cast);
-
-	if (cast->pwr_stream_state) {
-		logprint(DEBUG, "wlroots: request pipewire param change");
+	if (cast->frame_state == XDPW_FRAME_STATE_RENEG) {
 		pwr_update_stream_param(cast);
 	}
+
+	if (cast->frame_state == XDPW_FRAME_STATE_FAILED) {
+		xdpw_pwr_enqueue_buffer(cast);
+	}
+
+	if (cast->frame_state == XDPW_FRAME_STATE_SUCCESS) {
+		xdpw_pwr_enqueue_buffer(cast);
+		uint64_t delay_ns = fps_limit_measure_end(&cast->fps_limit, cast->framerate);
+		if (delay_ns > 0) {
+			xdpw_add_timer(cast->ctx->state, delay_ns,
+				(xdpw_event_loop_timer_func_t) xdpw_wlr_frame_start, cast);
+			return;
+		}
+	}
+	xdpw_wlr_frame_start(cast);
+}
+
+void xdpw_wlr_frame_start(struct xdpw_screencast_instance *cast) {
+	logprint(TRACE, "wlroots: start screencopy");
+	if (cast->quit || cast->err) {
+		xdpw_screencast_instance_destroy(cast);
+		return;
+	}
+
+	if (cast->initialized && !cast->pwr_stream_state) {
+		cast->frame_state = XDPW_FRAME_STATE_NONE;
+		return;
+	}
+
+	cast->frame_state = XDPW_FRAME_STATE_STARTED;
+	xdpw_wlr_register_cb(cast);
+}
+
+static void wlr_frame_buffer_done(void *data,
+		struct zwlr_screencopy_frame_v1 *frame);
+
+static void wlr_frame_buffer(void *data, struct zwlr_screencopy_frame_v1 *frame,
+		uint32_t format, uint32_t width, uint32_t height, uint32_t stride) {
+	struct xdpw_screencast_instance *cast = data;
+	if (!frame) {
+		return;
+	}
+
+	logprint(TRACE, "wlroots: buffer event handler");
+	cast->wlr_frame = frame;
+
+	cast->screencopy_frame_info[WL_SHM].width = width;
+	cast->screencopy_frame_info[WL_SHM].height = height;
+	cast->screencopy_frame_info[WL_SHM].stride = stride;
+	cast->screencopy_frame_info[WL_SHM].size = stride * height;
+	cast->screencopy_frame_info[WL_SHM].format = xdpw_format_drm_fourcc_from_wl_shm(format);
+
+	if (zwlr_screencopy_manager_v1_get_version(cast->ctx->screencopy_manager) < 3) {
+		wlr_frame_buffer_done(cast, frame);
+	}
 }
 
 static void wlr_frame_linux_dmabuf(void *data,
 		struct zwlr_screencopy_frame_v1 *frame,
 		uint32_t format, uint32_t width, uint32_t height) {
+	struct xdpw_screencast_instance *cast = data;
+	if (!frame) {
+		return;
+	}
+
 	logprint(TRACE, "wlroots: linux_dmabuf event handler");
+
+	cast->screencopy_frame_info[DMABUF].width = width;
+	cast->screencopy_frame_info[DMABUF].height = height;
+	cast->screencopy_frame_info[DMABUF].format = format;
 }
 
 static void wlr_frame_buffer_done(void *data,
 		struct zwlr_screencopy_frame_v1 *frame) {
 	struct xdpw_screencast_instance *cast = data;
+	if (!frame) {
+		return;
+	}
 
 	logprint(TRACE, "wlroots: buffer_done event handler");
 
-	zwlr_screencopy_frame_v1_copy_with_damage(frame, cast->simple_frame.buffer);
+	if (!cast->initialized) {
+		xdpw_wlr_frame_finish(cast);
+		return;
+	}
+
+	// Check if announced screencopy information is compatible with pipewire meta
+	if ((cast->pwr_format.format != xdpw_format_pw_from_drm_fourcc(cast->screencopy_frame_info[cast->buffer_type].format) &&
+			cast->pwr_format.format != xdpw_format_pw_strip_alpha(xdpw_format_pw_from_drm_fourcc(cast->screencopy_frame_info[cast->buffer_type].format))) ||
+			cast->pwr_format.size.width != cast->screencopy_frame_info[cast->buffer_type].width ||
+			cast->pwr_format.size.height != cast->screencopy_frame_info[cast->buffer_type].height) {
+		logprint(DEBUG, "wlroots: pipewire and wlroots metadata are incompatible. Renegotiate stream");
+		cast->frame_state = XDPW_FRAME_STATE_RENEG;
+		xdpw_wlr_frame_finish(cast);
+		return;
+	}
+
+	if (!cast->current_frame.xdpw_buffer) {
+		xdpw_pwr_dequeue_buffer(cast);
+	}
+
+	if (!cast->current_frame.xdpw_buffer) {
+		logprint(WARN, "wlroots: no current buffer");
+		xdpw_wlr_frame_finish(cast);
+		return;
+	}
+
+	assert(cast->current_frame.xdpw_buffer);
+
+	// Check if dequeued buffer is compatible with announced buffer
+	if (( cast->buffer_type == WL_SHM &&
+				(cast->current_frame.xdpw_buffer->size[0] != cast->screencopy_frame_info[cast->buffer_type].size ||
+				cast->current_frame.xdpw_buffer->stride[0] != cast->screencopy_frame_info[cast->buffer_type].stride)) ||
+			cast->current_frame.xdpw_buffer->width != cast->screencopy_frame_info[cast->buffer_type].width ||
+			cast->current_frame.xdpw_buffer->height != cast->screencopy_frame_info[cast->buffer_type].height) {
+		logprint(DEBUG, "wlroots: pipewire buffer has wrong dimensions");
+		cast->frame_state = XDPW_FRAME_STATE_FAILED;
+		xdpw_wlr_frame_finish(cast);
+		return;
+	}
+
+	zwlr_screencopy_frame_v1_copy_with_damage(frame, cast->current_frame.xdpw_buffer->buffer);
 	logprint(TRACE, "wlroots: frame copied");
 
 	fps_limit_measure_start(&cast->fps_limit, cast->framerate);
 }
 
-static void wlr_frame_buffer(void *data, struct zwlr_screencopy_frame_v1 *frame,
-		uint32_t format, uint32_t width, uint32_t height, uint32_t stride) {
-	struct xdpw_screencast_instance *cast = data;
-
-	logprint(TRACE, "wlroots: buffer event handler");
-	cast->wlr_frame = frame;
-	if (cast->simple_frame.width != width ||
-			cast->simple_frame.height != height ||
-			cast->simple_frame.stride != stride ||
-			cast->simple_frame.format != format) {
-		logprint(TRACE, "wlroots: buffer properties changed");
-		wlr_frame_buffer_chparam(cast, format, width, height, stride);
-	}
-
-	if (cast->simple_frame.buffer == NULL) {
-		logprint(DEBUG, "wlroots: create shm buffer");
-		cast->simple_frame.buffer = create_shm_buffer(cast, format, width, height,
-			stride, &cast->simple_frame.data);
-	} else {
-		logprint(TRACE,"wlroots: shm buffer exists");
-	}
-
-	if (cast->simple_frame.buffer == NULL) {
-		logprint(ERROR, "wlroots: failed to create buffer");
-		abort();
-	}
-
-	if (zwlr_screencopy_manager_v1_get_version(cast->ctx->screencopy_manager) < 3) {
-		wlr_frame_buffer_done(cast,frame);
-	}
-}
-
 static void wlr_frame_flags(void *data, struct zwlr_screencopy_frame_v1 *frame,
 		uint32_t flags) {
 	struct xdpw_screencast_instance *cast = data;
-
-	logprint(TRACE, "wlroots: flags event handler");
-	cast->simple_frame.y_invert = flags & ZWLR_SCREENCOPY_FRAME_V1_FLAGS_Y_INVERT;
-}
-
-static void wlr_frame_ready(void *data, struct zwlr_screencopy_frame_v1 *frame,
-		uint32_t tv_sec_hi, uint32_t tv_sec_lo, uint32_t tv_nsec) {
-	struct xdpw_screencast_instance *cast = data;
-
-	logprint(TRACE, "wlroots: ready event handler");
-
-	cast->simple_frame.tv_sec = ((((uint64_t)tv_sec_hi) << 32) | tv_sec_lo);
-	cast->simple_frame.tv_nsec = tv_nsec;
-
-	if (!cast->quit && !cast->err && cast->pwr_stream_state) {
-		pw_loop_signal_event(cast->ctx->state->pw_loop, cast->event);
+	if (!frame) {
 		return;
 	}
 
-	xdpw_wlr_frame_free(cast);
-}
-
-static void wlr_frame_failed(void *data,
-		struct zwlr_screencopy_frame_v1 *frame) {
-	struct xdpw_screencast_instance *cast = data;
-
-	logprint(TRACE, "wlroots: failed event handler");
-	cast->err = true;
-
-	xdpw_wlr_frame_free(cast);
+	logprint(TRACE, "wlroots: flags event handler");
+	cast->current_frame.y_invert = flags & ZWLR_SCREENCOPY_FRAME_V1_FLAGS_Y_INVERT;
 }
 
 static void wlr_frame_damage(void *data, struct zwlr_screencopy_frame_v1 *frame,
 		uint32_t x, uint32_t y, uint32_t width, uint32_t height) {
 	struct xdpw_screencast_instance *cast = data;
+	if (!frame) {
+		return;
+	}
 
 	logprint(TRACE, "wlroots: damage event handler");
 
-	cast->simple_frame.damage.x = x;
-	cast->simple_frame.damage.y = y;
-	cast->simple_frame.damage.width = width;
-	cast->simple_frame.damage.height = height;
+	cast->current_frame.damage.x = x;
+	cast->current_frame.damage.y = y;
+	cast->current_frame.damage.width = width;
+	cast->current_frame.damage.height = height;
+}
+
+static void wlr_frame_ready(void *data, struct zwlr_screencopy_frame_v1 *frame,
+		uint32_t tv_sec_hi, uint32_t tv_sec_lo, uint32_t tv_nsec) {
+	struct xdpw_screencast_instance *cast = data;
+	if (!frame) {
+		return;
+	}
+
+	logprint(TRACE, "wlroots: ready event handler");
+
+	cast->current_frame.tv_sec = ((((uint64_t)tv_sec_hi) << 32) | tv_sec_lo);
+	cast->current_frame.tv_nsec = tv_nsec;
+	logprint(TRACE, "wlroots: timestamp %"PRIu64":%"PRIu32, cast->current_frame.tv_sec, cast->current_frame.tv_nsec);
+
+	cast->frame_state = XDPW_FRAME_STATE_SUCCESS;
+
+	xdpw_wlr_frame_finish(cast);
+}
+
+static void wlr_frame_failed(void *data,
+		struct zwlr_screencopy_frame_v1 *frame) {
+	struct xdpw_screencast_instance *cast = data;
+	if (!frame) {
+		return;
+	}
+
+	logprint(TRACE, "wlroots: failed event handler");
+
+	cast->frame_state = XDPW_FRAME_STATE_FAILED;
+
+	xdpw_wlr_frame_finish(cast);
 }
 
 static const struct zwlr_screencopy_frame_v1_listener wlr_frame_listener = {
@@ -545,6 +571,167 @@ static void wlr_remove_output(struct xdpw_wlr_output *out) {
 	free(out);
 }
 
+static void wlr_format_modifier_pair_add(struct xdpw_screencast_context *ctx,
+		uint32_t format, uint64_t modifier) {
+	struct xdpw_format_modifier_pair *fm_pair;
+	wl_array_for_each(fm_pair, &ctx->format_modifier_pairs) {
+		if (fm_pair->fourcc == format && fm_pair->modifier == modifier) {
+			logprint(TRACE, "wlroots: skipping duplicated format %u (%lu)", fm_pair->fourcc, fm_pair->modifier);
+			return;
+		}
+	}
+
+	fm_pair = wl_array_add(&ctx->format_modifier_pairs, sizeof(struct xdpw_format_modifier_pair));
+	fm_pair->fourcc = format;
+	fm_pair->modifier = modifier;
+	logprint(TRACE, "wlroots: format %u (%lu)", fm_pair->fourcc, fm_pair->modifier);
+}
+
+static void linux_dmabuf_handle_modifier(void *data,
+		struct zwp_linux_dmabuf_v1 *zwp_linux_dmabuf_v1,
+		uint32_t format, uint32_t modifier_hi, uint32_t modifier_lo) {
+	struct xdpw_screencast_context *ctx = data;
+
+	logprint(TRACE, "wlroots: linux_dmabuf_handle_modifier called");
+
+	uint64_t modifier = (((uint64_t)modifier_hi) << 32) | modifier_lo;
+	wlr_format_modifier_pair_add(ctx, format, modifier);
+}
+
+static const struct zwp_linux_dmabuf_v1_listener linux_dmabuf_listener = {
+	.format = noop,
+	.modifier = linux_dmabuf_handle_modifier,
+};
+
+static void linux_dmabuf_feedback_handle_main_device(void *data,
+		struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1, struct wl_array *device_arr) {
+	struct xdpw_screencast_context *ctx = data;
+
+	logprint(DEBUG, "wlroots: linux_dmabuf_feedback_handle_main_device called");
+
+	assert(ctx->gbm == NULL);
+
+	dev_t device;
+	assert(device_arr->size == sizeof(device));
+	memcpy(&device, device_arr->data, sizeof(device));
+
+	drmDevice *drmDev;
+	if (drmGetDeviceFromDevId(device, /* flags */ 0, &drmDev) != 0) {
+		logprint(WARN, "wlroots: unable to open main device");
+		ctx->state->config->screencast_conf.force_mod_linear = true;
+		return;
+	}
+	ctx->gbm = xdpw_gbm_device_create(drmDev);
+}
+
+static void linux_dmabuf_feedback_format_table(void *data,
+		struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1, int fd, uint32_t size) {
+	struct xdpw_screencast_context *ctx = data;
+
+	logprint(DEBUG, "wlroots: linux_dmabuf_feedback_format_table called");
+
+	wl_array_release(&ctx->format_modifier_pairs);
+	wl_array_init(&ctx->format_modifier_pairs);
+
+	ctx->feedback_data.format_table_data = mmap(NULL , size, PROT_READ, MAP_PRIVATE, fd, 0);
+	if (ctx->feedback_data.format_table_data == MAP_FAILED) {
+		ctx->feedback_data.format_table_data = NULL;
+		ctx->feedback_data.format_table_size = 0;
+		return;
+	}
+	ctx->feedback_data.format_table_size = size;
+}
+
+static void linux_dmabuf_feedback_handle_done(void *data,
+		struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1) {
+	struct xdpw_screencast_context *ctx = data;
+
+	logprint(DEBUG, "wlroots: linux_dmabuf_feedback_handle_done called");
+
+	if (ctx->feedback_data.format_table_data) {
+		munmap(ctx->feedback_data.format_table_data, ctx->feedback_data.format_table_size);
+	}
+	ctx->feedback_data.format_table_data = NULL;
+	ctx->feedback_data.format_table_size = 0;
+}
+
+static void linux_dmabuf_feedback_tranche_target_devices(void *data,
+		struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1, struct wl_array *device_arr) {
+	struct xdpw_screencast_context *ctx = data;
+
+	logprint(DEBUG, "wlroots: linux_dmabuf_feedback_tranche_target_devices called");
+
+	dev_t device;
+	assert(device_arr->size == sizeof(device));
+	memcpy(&device, device_arr->data, sizeof(device));
+
+	drmDevice *drmDev;
+	if (drmGetDeviceFromDevId(device, /* flags */ 0, &drmDev) != 0) {
+		return;
+	}
+
+	if (ctx->gbm) {
+		drmDevice *drmDevRenderer = NULL;
+		drmGetDevice2(gbm_device_get_fd(ctx->gbm), /* flags */ 0, &drmDevRenderer);
+		ctx->feedback_data.device_used = drmDevicesEqual(drmDevRenderer, drmDev);
+	} else {
+		ctx->gbm = xdpw_gbm_device_create(drmDev);
+		ctx->feedback_data.device_used = ctx->gbm != NULL;
+	}
+}
+
+static void linux_dmabuf_feedback_tranche_flags(void *data,
+		struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1, uint32_t flags) {
+	logprint(DEBUG, "wlroots: linux_dmabuf_feedback_tranche_flags called");
+}
+
+static void linux_dmabuf_feedback_tranche_formats(void *data,
+		struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1, struct wl_array *indices) {
+	struct xdpw_screencast_context *ctx = data;
+
+	logprint(DEBUG, "wlroots: linux_dmabuf_feedback_tranche_formats called");
+
+	if (!ctx->feedback_data.device_used || !ctx->feedback_data.format_table_data) {
+		return;
+	}
+	struct fm_entry {
+		uint32_t format;
+		uint32_t padding;
+		uint64_t modifier;
+	};
+	// An entry in the table has to be 16 bytes long
+	assert(sizeof(struct fm_entry) == 16);
+
+	uint32_t n_modifiers = ctx->feedback_data.format_table_size/sizeof(struct fm_entry);
+	struct fm_entry *fm_entry = ctx->feedback_data.format_table_data;
+	uint16_t *idx;
+	wl_array_for_each(idx, indices) {
+		if (*idx >= n_modifiers) {
+			continue;
+		}
+		wlr_format_modifier_pair_add(ctx, (fm_entry + *idx)->format, (fm_entry + *idx)->modifier);
+	}
+}
+
+static void linux_dmabuf_feedback_tranche_done(void *data,
+		struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1) {
+	struct xdpw_screencast_context *ctx = data;
+
+	logprint(DEBUG, "wlroots: linux_dmabuf_feedback_tranche_done called");
+
+	ctx->feedback_data.device_used = false;
+}
+
+static const struct zwp_linux_dmabuf_feedback_v1_listener linux_dmabuf_listener_feedback = {
+	.main_device = linux_dmabuf_feedback_handle_main_device,
+	.format_table = linux_dmabuf_feedback_format_table,
+	.done = linux_dmabuf_feedback_handle_done,
+	.tranche_target_device = linux_dmabuf_feedback_tranche_target_devices,
+	.tranche_flags = linux_dmabuf_feedback_tranche_flags,
+	.tranche_formats = linux_dmabuf_feedback_tranche_formats,
+	.tranche_done = linux_dmabuf_feedback_tranche_done,
+};
+
 static void wlr_registry_handle_add(void *data, struct wl_registry *reg,
 		uint32_t id, const char *interface, uint32_t ver) {
 	struct xdpw_screencast_context *ctx = data;
@@ -586,6 +773,24 @@ static void wlr_registry_handle_add(void *data, struct wl_registry *reg,
 		ctx->xdg_output_manager =
 			wl_registry_bind(reg, id, &zxdg_output_manager_v1_interface, XDG_OUTPUT_MANAGER_VERSION);
 	}
+	if (strcmp(interface, zwp_linux_dmabuf_v1_interface.name) == 0) {
+		uint32_t version = ver;
+		if (LINUX_DMABUF_VERSION < ver) {
+			version = LINUX_DMABUF_VERSION;
+		} else if (LINUX_DMABUF_VERSION_MIN > ver) {
+			logprint(INFO, "wlroots: interface %s (Version %u) is required for DMA-BUF screencast", interface, LINUX_DMABUF_VERSION_MIN);
+			return;
+		}
+		logprint(DEBUG, "wlroots: |-- registered to interface %s (Version %u)", interface, version);
+		ctx->linux_dmabuf = wl_registry_bind(reg, id, &zwp_linux_dmabuf_v1_interface, version);
+
+		if (version >= 4) {
+			ctx->linux_dmabuf_feedback = zwp_linux_dmabuf_v1_get_default_feedback(ctx->linux_dmabuf);
+			zwp_linux_dmabuf_feedback_v1_add_listener(ctx->linux_dmabuf_feedback, &linux_dmabuf_listener_feedback, ctx);
+		} else {
+			zwp_linux_dmabuf_v1_add_listener(ctx->linux_dmabuf, &linux_dmabuf_listener, ctx);
+		}
+	}
 }
 
 static void wlr_registry_handle_remove(void *data, struct wl_registry *reg,
@@ -593,6 +798,23 @@ static void wlr_registry_handle_remove(void *data, struct wl_registry *reg,
 	struct xdpw_screencast_context *ctx = data;
 	struct xdpw_wlr_output *output = xdpw_wlr_output_find(ctx, NULL, id);
 	if (output) {
+		logprint(DEBUG, "wlroots: output removed (%s)", output->name);
+		struct xdpw_screencast_instance *cast, *tmp;
+		wl_list_for_each_safe(cast, tmp, &ctx->screencast_instances, link) {
+			if (cast->target_output == output) {
+				// screencopy might be in process for this instance
+				wlr_frame_free(cast);
+				// instance might be waiting for wakeup by the frame limiter
+				struct xdpw_timer *timer, *ttmp;
+				wl_list_for_each_safe(timer, ttmp, &cast->ctx->state->timers, link) {
+					if (timer->user_data == cast) {
+						xdpw_destroy_timer(timer);
+					}
+				}
+				cast->teardown = true;
+				xdpw_screencast_instance_teardown(cast);
+			}
+		}
 		wlr_remove_output(output);
 	}
 }
@@ -611,6 +833,9 @@ int xdpw_wlr_screencopy_init(struct xdpw_state *state) {
 	// initialize a list of active screencast instances
 	wl_list_init(&ctx->screencast_instances);
 
+	// initialize a list of format modifier pairs
+	wl_array_init(&ctx->format_modifier_pairs);
+
 	// retrieve registry
 	ctx->registry = wl_display_get_registry(state->wl_display);
 	wl_registry_add_listener(ctx->registry, &wlr_registry_listener, ctx);
@@ -645,10 +870,20 @@ int xdpw_wlr_screencopy_init(struct xdpw_state *state) {
 		return -1;
 	}
 
+	// make sure we have a gbm device
+	if (ctx->linux_dmabuf && !ctx->gbm) {
+		ctx->gbm = xdpw_gbm_device_create(NULL);
+		if (!ctx->gbm) {
+			logprint(ERROR, "System doesn't support gbm!");
+		}
+	}
+
 	return 0;
 }
 
 void xdpw_wlr_screencopy_finish(struct xdpw_screencast_context *ctx) {
+	wl_array_release(&ctx->format_modifier_pairs);
+
 	struct xdpw_wlr_output *output, *tmp_o;
 	wl_list_for_each_safe(output, tmp_o, &ctx->output_list, link) {
 		wl_list_remove(&output->link);
@@ -670,6 +905,17 @@ void xdpw_wlr_screencopy_finish(struct xdpw_screencast_context *ctx) {
 	if (ctx->xdg_output_manager) {
 		zxdg_output_manager_v1_destroy(ctx->xdg_output_manager);
 	}
+	if (ctx->gbm) {
+		int fd = gbm_device_get_fd(ctx->gbm);
+		gbm_device_destroy(ctx->gbm);
+		close(fd);
+	}
+	if (ctx->linux_dmabuf_feedback) {
+		zwp_linux_dmabuf_feedback_v1_destroy(ctx->linux_dmabuf_feedback);
+	}
+	if (ctx->linux_dmabuf) {
+		zwp_linux_dmabuf_v1_destroy(ctx->linux_dmabuf);
+	}
 	if (ctx->registry) {
 		wl_registry_destroy(ctx->registry);
 	}
